// module 433MhzGUI.cpp (Windows 32-Bit or Linux/GTK, uses wxWidgets)
// Copyright Dr.E.Huckert 02-2017
//
// A GUI for a remote 433Mhz based Audio player (Raspberry 3)
// (see Makefiles 433MhzGUI.mak or 433MhzGUIUnx.mak)
// Needs source modules inifilewx.cpp, hu_trace.cpp, util_hu.cpp

// Needs ini file 433MhzGUI.ini to find the serial device name
// The device name is normally "/dev/ttyUSB0" (on Linux)
// or "3" (on Windows) - but this may vary

// inifile sample (here for Windows):
// [general]
// device=3
// traceport=9514

// If you need UDP-trace output you may use my CLI program udp_server
//  (See on my web page www.huckert.com/ehuckert/huckert2.php under
//  the heading "UDP based logging for shell windows")

// versions
// 04.10.2017 created / derived from playEH.cpp (by EH)
// 07.10.2017 automatic scaling for the UI added - everything works fine
//            compiled/linked with wxWidgets V3.1.0 
//            (static libs, Unicode, 32Bit, Windows)
// 24.10.2017 fixed: problem garbage bytes after open() for /dev/ttyS0 (Linux only) 
//            Version V1.4
// 27.10.2017 initial message "HELLO433MhzGUI" added in OnInit()
//            final message "EXIT433MhzGUI" added, V1.5

#include <stdlib.h>
#include <time.h>
#include <sys/timeb.h>
//#include <unistd.h>
#include <stdlib.h>

#include "wx/wx.h"
#include "wx/string.h"
#include "wx/cursor.h"
#include "wx/list.h"
#include "wx/hashmap.h"

#include "hu_trace.h"
#include "inifilewx.h"

#include "433MhzGUI.xpm"               // an icon file in xpm format

#define WIDTH     400
#define HEIGTH    400
#define XPOS      (WIDTH/5)
#define YPOS      (HEIGTH/5)
#define LEDIT     160
#define LSTATIC   160
#define LBUTTON   120
#define CTLHEIGTH  26

wxString versionInfo =  "GUI433Mhz V1.5 Copyright Dr.Edgar Huckert 10-2017";

//
// global configuration
// UDP trace
// The port may be changed in ini file GUI433Mhz.ini (item "tracePort")
int        defaultTracePort      = 9514;
char       defaultTraceDest[128] = "localhost";
HU_TRACE   *pTrace = new HU_TRACE(::defaultTraceDest, ::defaultTracePort);

class GUI433MhzApp; // forward declaration
GUI433MhzApp *pApp        = NULL;
IniFileWx  *pIni          = NULL;
static int continueThread = 1;
static int globalErrors   = 0;
wxListBox  *pGlobalList   = NULL;

#ifdef WIN32
HANDLE  comm_handle = INVALID_HANDLE_VALUE;
wxString deviceName = "3";
#endif
#ifdef LINUX
// the device name may be changed via the ini file
wxString deviceName = "/dev/ttyUSB0";
int  serialFd       = -1;
#endif

// ----------------------------------------------------------
// sleep a number of milliseconds
int milliSleep(int millis)
{
  int ret;
  ret = 0;
#ifdef WIN32
  ::Sleep(millis);
#else
  ret = ::usleep((long)millis * 1000L);
#endif
  return ret;
}   // end milliSleep()

#ifdef WIN32
// ---------------------------------------
// Open a serial device (COMxx) under Windows
// Liefert: >= 0 if ok
//          < 0 if error
// side effect: comm_handle set 
int openSerialLine(int  comNo,
                   int  baudrate)
{
  int    ret = 0;
  HANDLE handle;
  DCB    dcb;
  DWORD  mask;
  wxString comm_device;
  //
  if (::pTrace != NULL) 
      ::pTrace->trace("openSerialLine() comNo=", comNo);
  if (comNo < 10)
    comm_device.Printf("COM%u", comNo);
  else
  {
    // s. Link: https://support.microsoft.com/en-us/kb/115831
    // COM ports >= 10 must be handled in a special way
    wxString aux;
    aux.Printf(wxT("COM%u"), comNo);
    comm_device.Printf("\\\\.\\");
    comm_device = comm_device  + aux;
  }
  if (::pTrace != NULL) 
      ::pTrace->trace("openSerialLine() deviceName=", deviceName);
  handle = ::CreateFile(comm_device.wc_str(), 
                        GENERIC_READ + GENERIC_WRITE,
                        0, 
                        0,
                        OPEN_EXISTING, 
  // for synchronous IO never use FILE_FLAG_OVERLAPPED here!
                        FILE_ATTRIBUTE_NORMAL, 
                        0);
  if (handle == INVALID_HANDLE_VALUE)
  {
    long code = ::GetLastError();
    if (::pTrace != NULL) 
      ::pTrace->trace("openSerialLine() error code=", code);
    ret = -2;
    goto zurueck;
  }
  // Open OK here
  ::comm_handle = handle;
  deviceName = comm_device;
  //
  memset((void *)&dcb, 0, sizeof(dcb));
  dcb.BaudRate      = (DWORD)baudrate;
  dcb.fBinary       = 1;
  dcb.fDtrControl   = DTR_CONTROL_DISABLE;
  dcb.fRtsControl   = RTS_CONTROL_DISABLE;
  dcb.fInX          = FALSE;
  dcb.fOutX         = FALSE;
  dcb.Parity        = NOPARITY;
  dcb.StopBits      = ONESTOPBIT;
  dcb.ByteSize      = 8;
  ::SetCommState(::comm_handle, &dcb);
  //
  mask = EV_RXCHAR;
  ::SetCommMask(::comm_handle, mask);
  /* from MS docu:
        A value of MAXDWORD, combined with zero values for both the ReadTotalTimeoutConstant 
         and ReadTotalTimeoutMultiplier members, specifies that the read operation is to return 
         immediately with the bytes that have already been received, even if no bytes have been received.
  */
  COMMTIMEOUTS  CommTimeOuts ;
  CommTimeOuts.ReadIntervalTimeout         = MAXDWORD;
  CommTimeOuts.ReadTotalTimeoutMultiplier  = 0;
  CommTimeOuts.ReadTotalTimeoutConstant    = 0;
  CommTimeOuts.WriteTotalTimeoutMultiplier = 1;
  CommTimeOuts.WriteTotalTimeoutConstant   = 2;
  if (! SetCommTimeouts(comm_handle, &CommTimeOuts))
  {
    ret = -3;
    goto zurueck;
  }
  ret = 0;
  //
  zurueck:
  if (::pTrace != NULL) 
      ::pTrace->trace("openSerialLine() ret=", ret);
  return ret;
}   // end openSerialLine()

// ----------------------------------------------- 
// Test if  data are available for read on a line
// Returns >0 if data are ready
// Returns < 0 upon error
int testDataAvailable()
{
  int   ret = 0;
  DWORD mask;
  //
  if (::comm_handle == INVALID_HANDLE_VALUE)
  {
    ret = -1;
    goto zurueck;
  }
  ::GetCommMask(::comm_handle, &mask);
  if (mask & EV_RXCHAR)
  {
    ret = 1;
    goto zurueck;
  }
  zurueck:
  return ret;
}   // end testDataAvailable()

// ----------------------------------------------- 
// Write a single character to the line (must be open)
// Note: 1-byte characters (ASCII, UTF-8) only!
// Returns < 0 upon error
int putSerialByte(const char ch)
{
  DWORD    nb;
  int      ret = 1;
  BOOL     retb;
  char     buf[48];
  //
  if (::comm_handle == INVALID_HANDLE_VALUE)
  {
    ret = -1;
    goto zurueck;
  }
  buf[0] = ch;
  retb = ::WriteFile(::comm_handle, 
                     (LPSTR)buf, 
                     (DWORD)1, 
                      &nb, 
                      NULL);
  if (::pTrace != NULL) 
      ::pTrace->trace("putSerialByte() retb,nb=", (int)retb, nb);
  if (! retb)
  {
    long code = ::GetLastError();
    if (::pTrace != NULL) 
      ::pTrace->trace("putSerialByte() error code=", code);
    ret = -2;
    goto zurueck;
  }
  if (nb != 1L) 
  {
    ret = -3;
    goto zurueck;
  }
  //
  zurueck:
  if (::pTrace != NULL) 
      ::pTrace->trace("putSerialByte() ret=", ret);
  return ret;
}   // end putSerialByte()

// ----------------------------------------------- 
// Windows
// Read a maximum of lbuf bytes from the line
// Returns: the number of bytes actually read
int readSerialBuf(char *buf, int lbuf)
{
  BOOL     ret;
  DWORD    nb = 0;
  char     ch;
  int      state;
  int      bufPos = 0;
  DWORD    mask;
  //
  if (::comm_handle == INVALID_HANDLE_VALUE) 
  {
    nb = -1;
    goto zurueck;
  } 
  while (lbuf > 0)
  {
    state = testDataAvailable();
    if (state > 0)
    {
      ret = ::ReadFile(::comm_handle, 
                        (LPSTR)&ch, 
                        (DWORD)1, 
                        &nb, 
                        NULL);  
      if (! ret)
      {
        nb = -2;
        goto zurueck;
      }
      if (nb > 0)  // fix 09-2017
      {
        buf[bufPos++] = ch;
        // Zeilenende?
        if ((ch & 0xff) < ' ')
          break;
        lbuf--;
      }
      else
        // wait 1 ms - this is the important wait!
        // This brings CPU performance really down
        ::Sleep(1L);
    }   // end if state > 0
    if (state == 0)
    {
      // wait 1 ms - not so important!
      ::Sleep(1L);
    }
    if (state < 0)
    {
      nb = -2;
      goto zurueck;
    }
  }   // end while ... 
  //
  nb = bufPos;
  zurueck:
  if (::pTrace != NULL) 
      ::pTrace->trace("readSerialBuf() ret=", nb);
  return (int)nb;
}   // end readSerialBuf()

// ----------------------------------------------- 
// Windows!
// Returns: < 0 if error
int closeSerialLine()
{
  int ret = 0;
  if (::comm_handle == INVALID_HANDLE_VALUE)
  {
    ret = -1;
    goto zurueck;
  }
  ::CloseHandle(::comm_handle);
  ::comm_handle = INVALID_HANDLE_VALUE;
  //
  zurueck:
  if (::pTrace != NULL) 
      ::pTrace->trace("closeSerialLine() ret=", ret);
  return ret;
}   // end closeSerialLine()
#endif

#ifdef LINUX
// ---------------------------------
// Open the serial line device for write
int openSerialLine(const char *pDevice)
{
  int fd, ret;
  char cmd[128];
  //udp_server >xx
  // open now
  fd = ::open(pDevice,  O_RDWR | O_NOCTTY | O_SYNC | O_TRUNC);
  if (::pTrace != NULL) 
      ::pTrace->trace("openSerialLine() device,fd=", pDevice, fd);
  if (fd < 0)
  {
    goto zurueck;
  }
  //
  // configure the device
  ::sprintf(cmd, "stty -F %s 9600 raw -echo -icanon", pDevice);
  ret = ::system(cmd);
  if (::pTrace != NULL) 
      ::pTrace->trace("openSerialLine() system(stty..) ret=", ret);
  //
  zurueck:
  return fd;
}  // end openSerialLine()

// ----------------------------------------------- 
// Returns: < 0 if error
int closeSerialLine()
{
  int ret = 0;
  //
  if (::serialFd >= 0)
    ::close(::serialFd);
  else
    ret = -1;
  return ret;
}  // end closeSerialLine()

// ----------------------------------------------- 
// Write a single character to the line (must be open)
// Note: 1-byte characters (ASCII, UTF-8) only!
// Returns < 0 upon error
int putSerialByte(const char ch)
{
  int      ret = 1;
  int      nb = 0;
  static char sendBuf[1];
  //
  if (::serialFd < 0)
  {
    ret = -1;
    goto zurueck;
  }
  sendBuf[0] = ch;
  nb = ::write(::serialFd, sendBuf, 1);
  if (nb < 0)
  {
    // write error
    ret = -2;
    goto zurueck;
  }
  if (nb == 0)
  {
    // line has been closed
    ret = -3;
    goto zurueck;
  }
  //
  zurueck:
  if (::pTrace != NULL) 
      ::pTrace->trace("putSerialByte() ret=", ret);
  return ret;
}   // end putSerialByte()
#endif

// ------------------------------------------------------------------
// Windows and Linux
// parameter string s should be terminated by CR and LF
// We write single chars as we can delay them if necessary
// Returns < 0 upon error
int writeSerialLine(const wxString & s)
{
  int ret = 0;
  int nb = 0;
  int totalNb = 0;
  unsigned char ch;
  wxScopedCharTypeBuffer<char> cbuf;  // note: 1 byte per character
  //
  if (::pTrace != NULL) 
      ::pTrace->trace("writeSerialLine() buffer=", s);
  // convert parameter s to a 1-byte string
  cbuf = s.utf8_str();
  //
  // send now on a 1-byte base - delaying slightly each byte
  for (int n=0; n < s.Length(); n++)
  {
    milliSleep(1);
    ch = (unsigned char)(cbuf[n]);
    nb = ::putSerialByte(ch);
    if (nb < 0)
    {
      ret = -1;
      goto zurueck;
    }
    if (nb == 0)
    {
      ret = -2;
      goto zurueck;
    }
    totalNb += nb;
  }   //end for n
  //
  ret = totalNb;
  //
  zurueck:
  if (::pTrace != NULL) 
      ::pTrace->trace("writeSerialLine() ret=", ret);
  return ret;
}   // end writeSerialLine()

// ------------------------------------------------------------
class GUI433MhzThread : public wxThread
{
  public:
  GUI433MhzThread();
  void * Entry();
};

// -------------------------------------------------------------------
// Eliminates unwanted characters from the buffer
void cleanBuffer(char *buf)
{
 int n;
 if (buf == NULL)
   return;
 for (n=0; buf[n] != 0; n++)
 {
   if (buf[n] == '\n') buf[n] = 0;
   if (buf[n] == '\r') buf[n] = 0;
 }
}   // end clean_buffer()

// ------------------------------------------------------------------
// sets glob.var. globalErrors if error
// Returns an error message if an erroc is detected
// Returns an empty string if OK
 wxString initialize_433Mhz()
{
  int status = 0;
  wxString retMsg = "";
  //
  #ifdef WIN32
  long comNo = 0L;
  ::deviceName.ToLong(&comNo);
  status = ::openSerialLine((int)comNo, 9600);
  if (status < 0)
  {
    ::globalErrors++;
     if (::pTrace != NULL) 
      ::pTrace->trace("ERROR initialize_433Mhz(): cannot open device=",
                       ::deviceName);
     retMsg = "ERROR: cannot open device=";
     retMsg = retMsg + ::deviceName;
  }
  else
  {
    if (::pTrace != NULL) 
      ::pTrace->trace("initialize_433Mhz(): open OK for device=",
                      ::deviceName);
  }
  #endif
  #ifdef LINUX
  ::serialFd = ::openSerialLine(::deviceName);
  if (::serialFd < 0)
  {
    ::globalErrors++;
     if (::pTrace != NULL) 
      ::pTrace->trace("ERROR initialize_433Mhz(): cannot open device=",
                       ::deviceName);
     retMsg = "ERROR: cannot open device=";
     retMsg = retMsg + ::deviceName;
  }
  else
  {
    if (::pTrace != NULL) 
      ::pTrace->trace("initialize_433Mhz(): open OK for device=",
                      ::deviceName);
  }
  #endif
  return retMsg;
}   // end initialize_433Mhz()
  
// ------------------------------------------------------------------
// constructor
GUI433MhzThread::GUI433MhzThread()
{
}

// -------------------------------------------------------------------
// Thread procedure 
// Not really needed -we keept it to have some thread if we need one
void * GUI433MhzThread::Entry()
{
  static unsigned long ret = 0;
  if (::pTrace != NULL) 
    ::pTrace->trace("udp_receiver() GUI433MhzThread started");
  while (::continueThread)
  {
    ::milliSleep(100);
  }
  if (::pTrace != NULL) 
    ::pTrace->trace("udp_receiver() GUI433MhzThread finished");
  return (void *)&ret;
}   // end GUI433MhzThread::Entry()

// -----------------------------------------------------------
// Define a new application type
class GUI433MhzApp : public wxApp
{
  private:
  int tracePort;
public:
  GUI433MhzApp()
  { this->tracePort = -1; };
  virtual bool OnInit();
  void  getScaling(int *w, int *h);
  void OnAssertFailure(const wxChar * file,
                       int  	line,
                       const wxChar *  	func,
                       const wxChar *  	cond,
                       const wxChar *  	msg);
  //bool ProcessEvent(wxEvent & event);
};

// -------------------------------------------------
// This is the windows where we draw on
class GUI433MhzWindow : public wxScrolledWindow
{
  public:
    GUI433MhzWindow(wxPanel *, long, wxPoint, wxSize);
    void draw();
    int getDrawStatus() { return this->isDrawn; };
  private:
    wxWindowDC *pdc;
    int isDrawn;
};   // end class GUI433MhzWindow

// ------------------------------------------------------------
class GUI433MhzFrame : public wxFrame
{
public:
  GUI433MhzFrame(int xp, int yp, int xl, int yl);
  ~GUI433MhzFrame();

  void  OnQuit(wxCommandEvent& event);
  void  OnClose(wxCloseEvent& event);
  void  OnAbout(wxCommandEvent& event);
  void  OnButton(wxCommandEvent& event);
  void  OnSetFocus(wxFocusEvent& event);
  void  OnActivate(wxActivateEvent& event);
  void  OnMove(wxMoveEvent & event);
  void  OnSize(wxSizeEvent & event);
private:
  int terminate();
  GUI433MhzThread      *serverThread;
  wxListBox         *pListbox;
  wxPanel           *panel;
  wxMenu            *m_menuFile;
  wxMenuBar         *m_menuBar;
  wxColour          bgColour;
  wxColour          colour1;
  wxColour          colour2;
  // the song buttons
  wxButton          *pButton1;
  wxButton          *pButton2;
  wxButton          *pButton3;
  wxButton          *pButton4;
  wxButton          *pButton5;
  wxButton          *pButton6;
  wxButton          *pButton7;
  wxButton          *pButton8;
  wxButton          *pButton9;
  wxButton          *pButtonVL1;
  wxButton          *pButtonVL2;
  wxButton          *pButtonVM1;
  wxButton          *pButtonVM2;
  //
  wxButton          *pButtonStop;
  wxCursor          *stdCursor;
  //
  void              makePanel();
  wxStaticText *makeStaticText(const char *pLabel, 
                                int startx,
                                int starty,
                                int lctl, 
                                int hctl);
  wxTextCtrl   *makeEditField(const char *label,
                            int startx,
                            int starty,
                            int lstatic,
                            int lctl, 
                            int hctl,
                            long style);

  DECLARE_EVENT_TABLE()
};

enum
{
  BUTTON1       = 11000,
  BUTTON2       = 11001,
  BUTTON3       = 11002,
  BUTTON4       = 11003,
  BUTTON5       = 11004,
  BUTTON6       = 11005,
  BUTTON7       = 11006,
  BUTTON8       = 11007,
  BUTTON9       = 11008,
  BUTTONVolL1   = 11010,
  BUTTONVolM1   = 11011,
  BUTTONVolL2   = 11012,
  BUTTONVolM2   = 11013,		
  BUTTONStop    = 11020,
  SERVER_ABOUT  = wxID_ABOUT
};

BEGIN_EVENT_TABLE(GUI433MhzFrame, wxFrame)
  EVT_MENU(wxID_EXIT,   GUI433MhzFrame::OnQuit)
  EVT_CLOSE(GUI433MhzFrame::OnClose)
  EVT_MENU(SERVER_ABOUT,  GUI433MhzFrame::OnAbout)
  EVT_SET_FOCUS(GUI433MhzFrame::OnSetFocus)
  EVT_ACTIVATE(GUI433MhzFrame::OnActivate)
  //EVT_CLOSE(GUI433MhzFrame::OnQuit)              // TODO: wieder rein!
  EVT_MOVE(GUI433MhzFrame::OnMove)
  EVT_BUTTON(BUTTON1, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTON2, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTON3, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTON4, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTON5, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTON6, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTON7, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTON8, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTON9, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTONVolL1, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTONVolM1, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTONVolL2, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTONVolM2, GUI433MhzFrame::OnButton)
  EVT_BUTTON(BUTTONStop,  GUI433MhzFrame::OnButton)
  //EVT_SIZE(GUI433MhzFrame::OnSize)  // nicht aktivieren: Fehler!!!
END_EVENT_TABLE()

IMPLEMENT_APP(GUI433MhzApp)

// ---------------------------------------------------------------------
// Constructor
GUI433MhzWindow::GUI433MhzWindow(wxPanel *parent,   // GUI433MhzFrame *parent, 
                         long style, 
                         wxPoint position, 
                         wxSize size) 
    : wxScrolledWindow((wxFrame *)parent, -1, position, size, style)
{
  this->pdc     = NULL;
  this->isDrawn = 0;
  wxColour col("YELLOW");
  this->SetBackgroundColour(col);
  this->Show();
}   // end class GUI433MhzWindow

// ----------------------------------------------------------
int GUI433MhzFrame::terminate()
{
  int ret = 0;
  // terminate the UDP receiver thread
  ::continueThread = 0;
  // wait some milliseconds so that the UDP receiver thread
  // can terminate
  ::milliSleep(500);
  //
  if (this->serverThread != NULL)
  {
    this->serverThread = NULL;
  }
  //
  //sending final message "EXIT433MhzGUI"
  // This message should be ignored in the server
  #ifdef LINUX
  if (::serialFd >= 0)
  #else
  if (::comm_handle != INVALID_HANDLE_VALUE) 
  #endif
  {
    wxString s = "EXIT433MhzGUI\r\n";
    if (::pTrace != NULL) 
      ::pTrace->trace("GUI433MhzFrame::terminate() sending:", s);
    int status = ::writeSerialLine(s);
    closeSerialLine();
  }
  return ret;
}   // end GUI433MhzFrame::terminate()

// -------------------------------------------------------------
void GUI433MhzWindow::draw()
{
  //if (::pTrace != NULL) ::pTrace->trace("GUI433MhzWindow::draw()");
}


// ----------------------------------------------------------------
// noetig wegen obskurer CURSOR Probleme unter wxWidgets V3.0.1
void GUI433MhzApp::OnAssertFailure(const wxChar * file,
                                    int  	line,
                                    const wxChar *  	func,
                                    const wxChar *  	cond,
                                    const wxChar *  	msg) 	
{
  if (::pTrace != NULL) 
  {
    ::pTrace->trace("MyFrame::OnAssertFailure() file=", file);
    ::pTrace->trace("MyFrame::OnAssertFailure() line=", line);
    ::pTrace->trace("MyFrame::OnAssertFailure() func=", func);
  }
}   // end GUI433MhzApp::OnAssertFailure()

// ------------------------------------------------------------------
// get the appropriate values for x and y depending on 
// the screen resolution
void GUI433MhzApp::getScaling(int *width, int *heigth)
{
    //
  // get the x and y resolution of the screen
  int xres = wxSystemSettings::GetMetric(wxSYS_SCREEN_X);
  int yres = wxSystemSettings::GetMetric(wxSYS_SCREEN_Y);
  if (::pTrace != NULL) 
  {
    ::pTrace->trace("MyFrame::GetScaling() xres,yres=", xres,yres);
  }
  // calculate the optimal size
  if (xres > 640)
  {
    if (xres > 1000)
      *width = *width * 2.0;
    else
      *width = *width * 1.75;
  }
  if (yres > 400)
  {
    if (yres > 800)
      *heigth = *heigth * 2.0;
    else
      *heigth = *heigth * 1.75;
  }
}   // end GUI433MhzApp::getScaling()

// -------------------------------------------------------------------
// frame constructor
GUI433MhzFrame::GUI433MhzFrame(int xp, int yp, 
                               int xl, int yl) 
    : wxFrame((wxFrame *)NULL, 
              wxID_ANY,
              versionInfo,
              wxPoint(xp, yp), 
              wxSize(xl, yl))
{
  this->serverThread    = NULL;
  this->stdCursor       = new wxCursor(wxCURSOR_ARROW);
  this->pListbox        = NULL;
  
  this->bgColour = wxColour(200, 200, 200);
  this->colour1  = wxColour(64, 255, 64);
  this->colour2  = wxColour(255, 64, 64);

  // Make menus
  m_menuFile = new wxMenu();
  m_menuFile->Append(SERVER_ABOUT, 
                     _("&About...\tCtrl-A"), 
                     _("Show about dialog"));
  m_menuFile->AppendSeparator();
  m_menuFile->Append(wxID_EXIT, 
                    _("E&xit\tAlt-X"), 
                    _("Quit"));

  // Append menus to the menu bar
  m_menuBar = new wxMenuBar();
  m_menuBar->Append(m_menuFile, _("&File"));
  SetMenuBar(m_menuBar);
  //
  this->makePanel();
  // 
  // create the GUI433MhzThread thread and run it
  this->serverThread = new GUI433MhzThread();
  this->serverThread->Create();
  if (! (this->serverThread)->IsRunning())
  {
    wxThreadError err;
    err = (this->serverThread)->Run();
    if (err != wxTHREAD_NO_ERROR)
    {
      wxMessageBox("Fehler", _("cannot start thread"),
               wxOK | wxICON_INFORMATION, this);
    }
  }
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::Constructor leaving");
}   // end GUI433MhzFrame::GUI433MhzFrame()

// ----------------------------------------------------------------
// Destructor
// Assumes that other terminate actions are handled elsewhere
GUI433MhzFrame::~GUI433MhzFrame()
{
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::destructor");
  // as OnQuit() is not correctly called...
  // we terminate the thread here
  ::continueThread = 0;
  ::milliSleep(500);
  //
  if (::pTrace != NULL) 
    delete pTrace;
  pTrace = NULL;
}  // end destructor

// ------------------------------------------------------------------
// create a static text control, i.e. a label
wxStaticText *GUI433MhzFrame::makeStaticText(const char *pLabel, 
                                            int startx,
                                            int starty,
                                            int lstatic,
                                            int ctlheigth)
{
  wxStaticText *pTextCtl;
  pTextCtl = new wxStaticText(this->panel, 
                              -1, 
                              pLabel,
                              wxPoint(startx, starty),
                              wxSize(lstatic, ctlheigth));
  pTextCtl->SetBackgroundColour(this->bgColour);
  pTextCtl->Refresh();
  return pTextCtl;
}   // end GUI433MhzFrame::makeStaticText()

// -------------------------------------------------------------------
// Create a label and an edit field
// Return a pointer to the edit control
wxTextCtrl * GUI433MhzFrame::makeEditField(const char *label,
                                          int startx,
                                          int starty,
                                          int lstatic,
                                          int ledit,
                                          int ctlheigth,
                                          long style)
{
  wxStaticText *pTextCtl;
  wxTextCtrl   *pEditCtl;
  //
  pTextCtl = this->makeStaticText(label, startx, starty, 
                                  lstatic, ctlheigth);
  pEditCtl = new wxTextCtrl(this->panel, 
                            -1, 
                            "", 
                            wxPoint(startx + lstatic + 5, starty),
                            wxSize(ledit, ctlheigth),
                            style);
  //pEditCtl->SetEditable(true);
  return pEditCtl;
}   // end GUI433MhzFrame::makeEditField()

// ----------------------------------------------------------------
// create a panel i.e. the set of visible controls in the 
// main application window
void GUI433MhzFrame::makePanel()
{
  //if (::pTrace != NULL) ::pTrace->trace("GUI433MhzFrame::makePanel()");
  char aux[128];
  this->panel  = new wxPanel(this);
  int xp = 5;
  int yp = 5 ;
  int xl;
  int yl; 
  int lButton    = LBUTTON;
  int ctlHeigth  = CTLHEIGTH;
  int yDist      = 5;
  //
  // get the size of the panel in pixels
  this->GetClientSize(&xl, &yl); 
  ::pApp->getScaling(&xp, &yDist);
  ::pApp->getScaling(&lButton, &ctlHeigth);
  int leftMargin = 3 * xp;
  // 
  // the song buttons
  yp = yp + yDist; 
  this->pButton1 = new wxButton(this->panel, 
                                BUTTON1,
                                "W1",
                                wxPoint(xp + (0 * lButton) + leftMargin, yp),
                                wxSize(lButton, ctlHeigth));
  pButton1->SetBackgroundColour(this->colour1);
  this->pButton2 = new wxButton(this->panel, 
                                BUTTON2,
                                "W2",
                                wxPoint(xp + (1 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButton2->SetBackgroundColour(this->colour1);
  this->pButton3 = new wxButton(this->panel, 
                                BUTTON3,
                                "W3",
                                wxPoint(xp + (2 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButton3->SetBackgroundColour(this->colour1);
  yp = yp + ctlHeigth + yDist; 
  this->pButton4 = new wxButton(this->panel, 
                                BUTTON4,
                                "W4",
                                wxPoint(xp + (0 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButton4->SetBackgroundColour(this->colour1);
  this->pButton5 = new wxButton(this->panel, 
                                BUTTON5,
                                "W5",
                                wxPoint(xp + (1 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButton5->SetBackgroundColour(this->colour1);
  this->pButton6 = new wxButton(this->panel, 
                                BUTTON6,
                                "W6",
                                wxPoint(xp + (2 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButton6->SetBackgroundColour(this->colour1);
  yp = yp + ctlHeigth + yDist; 
  this->pButton7 = new wxButton(this->panel, 
                                BUTTON7,
                                "W7",
                                wxPoint(xp + (0 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButton7->SetBackgroundColour(this->colour1);
  this->pButton8 = new wxButton(this->panel, 
                                BUTTON8,
                                "W8",
                                wxPoint(xp + (1 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButton8->SetBackgroundColour(this->colour1);
  this->pButton9 = new wxButton(this->panel, 
                                BUTTON9,
                                "W9",
                                wxPoint(xp + (2 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButton9->SetBackgroundColour(this->colour1);
  //
  // Volume controls
  yp = yp + ctlHeigth + yDist; 
  this->pButtonVL1 = new wxButton(this->panel, 
                                BUTTONVolL1,
                                "Vol-",
                                wxPoint(xp + (0 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButtonVL1->SetBackgroundColour(this->colour1);
  this->pButtonVL2 = new wxButton(this->panel, 
                                BUTTONVolL2,
                                "Vol--",
                                wxPoint(xp + (1 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButtonVL2->SetBackgroundColour(this->colour1); 
  yp = yp + ctlHeigth + yDist; 
  this->pButtonVM1 = new wxButton(this->panel, 
                                BUTTONVolM1,
                                "Vol+",
                                wxPoint(xp + (0 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButtonVM1->SetBackgroundColour(this->colour1);
  this->pButtonVM2 = new wxButton(this->panel, 
                                BUTTONVolM2,
                                "Vol++",
                                wxPoint(xp + (1 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButtonVM2->SetBackgroundColour(this->colour1);
  //
  // the >STOP button (stop the remote player)
  this->pButtonStop = new wxButton(this->panel, 
                                BUTTONStop,
                                "STOP",
                                wxPoint(xp + (2 * lButton) + leftMargin, yp),
                               wxSize(lButton, ctlHeigth));
  pButtonStop->SetBackgroundColour(this->colour2);
  //
  // create a listbox where to show the UDP messages
  yp = yp + ctlHeigth + (3 * yDist); 
  this->pListbox = new wxListBox(this->panel, 
                                  wxSUNKEN_BORDER, 
                                  wxPoint(xp, yp),
                                  wxSize(xl - 15, yl - yp - 20));
  ::pGlobalList = this->pListbox;
}   // end GUI433MhzFrame::makePanel()

// ----------------------------------------------------------------
void GUI433MhzFrame::OnSetFocus(wxFocusEvent& event)
{
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::OnSetFocus()");
  //this->pw->draw(); 
}

// ----------------------------------------------------------------
void GUI433MhzFrame::OnActivate(wxActivateEvent& event)
{
  //~ if (this->pw != NULL) 
    //~ this->pw->draw(); 
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::OnActivate() id=", event.GetId());
}

// ----------------------------------------------------------------
void GUI433MhzFrame::OnMove(wxMoveEvent& event)
{
  // is too frequent
  //if (::pTrace != NULL) 
  //  ::pTrace->trace("GUI433MhzFrame::OnMove()");
  // get the actual position of the top left corner
  wxPoint pos = event.GetPosition();
  //if (::pTrace != NULL) ::pTrace->trace("GUI433MhzFrame::OnMove() x=", pos.x);
  //if (::pTrace != NULL) ::pTrace->trace("GUI433MhzFrame::OnMove() y=", pos.y);
  int xl, yl;
  // get the size of the window
  this->GetVirtualSize(&xl, &yl);
  //if (::pTrace != NULL) ::pTrace->trace("GUI433MhzFrame::OnMove() size x=", xl);
  //if (::pTrace != NULL) ::pTrace->trace("GUI433MhzFrame::OnMove() size y=", yl);
  //
  // check if the window is completely visible - if not we have to redraw the window!
  bool exp = this->IsExposed(pos.x, pos.y);
  //
  // redraw the inner window if it is not exposed
  if (! exp)
  {
    //~ if (this->pw != NULL)
    //~ {
      //~ if (this->pw->getDrawStatus()) 
        //~ this->pw->draw();
    //~ }
  }
}   // GUI433MhzFrame::OnMove()

// ----------------------------------------------------------------
// Vorsicht: wenn aktiviert anderes (problematisches) Verhalten! 
void GUI433MhzFrame::OnSize(wxSizeEvent& event)
{
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::OnSize()");
  //wxSize size = event.GetSize();
  //if (::pTrace != NULL) ::pTrace->trace("GUI433MhzFrame::OnSize() width=", size.GetWidth());
  //if (::pTrace != NULL) ::pTrace->trace("GUI433MhzFrame::OnSize() height=", size.GetHeight());
}

// ----------------------------------------------------------------
// Called when the X field in the window system pane is activated
void GUI433MhzFrame::OnClose(wxCloseEvent& event)
{
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::OnClose()");
  this->terminate();
  this->Destroy();
}   // end GUI433MhzFrame::OnClose()

// -------------------------------------------------------------
// Called when Exit is activated in the File menu
void GUI433MhzFrame::OnQuit(wxCommandEvent& event)
{
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::OnQuit()");
  this->terminate();
  this->Destroy();
}   // end GUI433MhzFrame::OnQuit()

// ----------------------------------------------------------------
void GUI433MhzFrame::OnAbout(wxCommandEvent& WXUNUSED(event))
{
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::OnAbout()");
  wxString s;
  s = "433Mhz controlled audio player\n";
  s = s + versionInfo;
  wxMessageBox(s, _("About GUI433Mhz"),
               wxOK | wxICON_INFORMATION, this);
}   // end GUI433MhzFrame::OnAbout()

// ------------------------------------------------------------------
// Is invoked when the application starts
bool GUI433MhzApp::OnInit()
{
  long cRet = 0;
  wxString s;
  //
  if (::pTrace != NULL)
    ::pTrace->trace("GUI433MhzApp::OnInit() start");
  //
  ::pApp = this;
  //
  // read the ini file GUI433Mhz.ini
  ::pIni = new IniFileWx();
  ::pIni->setLogging(::pTrace);
  #ifdef WIN32
  cRet = ::pIni->load(".\\433MhzGUI.ini");
  #else
  cRet = ::pIni->load("./433MhzGUI.ini");
  #endif
  if (::pTrace != NULL)
    ::pTrace->trace("GUI433MhzApp::OnInit() IniFileWx::load() ret=", cRet);
  // the trace port may have changed
  this->tracePort = ::pIni->getInt("general", "tracePort");
  if (this->tracePort >= 0)
  {
    if (::pTrace != NULL)
    {
      delete ::pTrace;
      ::pTrace = NULL;
    }
    ::pTrace = new HU_TRACE(::defaultTraceDest, this->tracePort);
    if (::pTrace != NULL)
      ::pTrace->trace("\n----------------------\n");
  }
  // this is for Linux: "/dev/ttyUSB0" or similar
  //         for Windows: "COM3:" or similar
  s = ::pIni->get("general", "device");
  if (s.length() > 0)
  {
    ::deviceName = s;
    if (::pTrace != NULL)
      ::pTrace->trace("GUI433MhzApp::OnInit() deviceName=", 
                      ::deviceName);
  }
  int xpos   = XPOS;
  int ypos   = YPOS;
  int width  = WIDTH;
  int heigth = HEIGTH;
  //
  // get the optimal size depending on the screen resolution
  ::pApp->getScaling(&width, &heigth);
  //
  // Create the main application window
  GUI433MhzFrame *pFrame = new GUI433MhzFrame(xpos, ypos, 
                                              width, heigth);
  pFrame->Show(true);
  SetTopWindow(pFrame); 
  //
  if (::pTrace != NULL)
      ::pTrace->trace("OnInit(): initializing 433Mhz communication");
  wxString errMsg = initialize_433Mhz();
  if (errMsg.Length() > 0)
  {
    if (::pTrace != NULL)
      ::pTrace->trace("probably 433Mhz comm. errors found - we terminate");
    ::wxMessageBox(errMsg);
    ::exit(-2);
  }
  //
  //sending initial message "HELLO433MhzGUI"
  // This message should be ignored in the server
  s = "HELLO433MhzGUI\r\n";
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::OnInit() sending:", s);
  int status = ::writeSerialLine(s);
  return true;
}   // end GUI433MhzApp::OnInit()

// ----------------------------------------------------------------
// The user has pressed a song button
// We send here messages over a serial line to a HC-12 module (433Mhz)
void GUI433MhzFrame::OnButton(wxCommandEvent& event)
{
  int ret = -1;
  int status = 0;
  wxString s;
  static int count = 0;
  //
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::OnButton() id=", event.GetId());
  wxButton *pButt = (wxButton *)(event.GetEventObject());
  if (pButt != NULL)
  {
    s = pButt->GetLabel();
    if (::pTrace != NULL) 
      ::pTrace->trace("GUI433MhzFrame::OnButton() label=", s);
    s = s + "\r\n";
    //
    //  write the button label on the serial line
    status = ::writeSerialLine(s);
    //
    if (::pTrace != NULL) 
      ::pTrace->trace("GUI433MhzFrame::OnButton() write() status=", status);
    if(status < 0 ) 
    {
      ::globalErrors++;
      ret = -2;
      goto zurueck;
    }
    else
    {
      // add the button label in the listbox (kind of protocol)
      wxString aux = "TX: " + s;
      aux = aux.Trim();
      if (::pGlobalList != NULL)
        ::pGlobalList->InsertItems(1, &aux, 0);
      ret = 0;
    }
  }
  zurueck:
  if (::pTrace != NULL) 
    ::pTrace->trace("GUI433MhzFrame::OnButton() ret=", ret);
}   // GUI433MhzFrame::OnButton()
