// module 433Mhz_server.c
// This is a 433Mhz based Audio server running an Raspberry 3
// It accepts commands (song names, volume commands, stop commands) 
// and handles them by starting appropriate actions.
// sample action: 
//   start a backgroud player ("playw")
// Special commands:
//    "STOP" terminates this server
//    "Vol+" increments the player volume (uses the amixer command) 
//    "Vol-" decrements the player volume (uses the amixer command) 
// Can handle only one 433Mhz connection!
// Written by Dr. E.Huckert 09-2017

// Notes:
//   getty or agetty processes must be stopped for the serial line!
//   No serial line attributes are set here - this must be done 
//   externally (ex.: stty -F /dev/ttyS0 9600 raw).
//   This program must run with root rights as normally the serial
//   device cannot be read or written by normal users.
//   Can handle only one 433Mhz connection!.

// compile/link:
//   gcc -o 433Mhz_server 433Mhz_server.c 

// Versions:
//   2017-10-01 first working version
//   2017-10-13 timeout added in readCommand()
//   2017-10-27 "HELLO" and "EXIT" messages ignored, V1.4

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <sys/timeb.h>

char exe_dir[]      = "/home/huckert/bin"; 
char play_cmd[]     = "playw";
char wave_dir[]     = "/home/huckert/Wave";
int logging         = 0;

// Note: the name for the serial device may be different 
// in other distributions or other machines
// This is for Raspbian V8.0("jessie")
// Can b changed in command line parameter "-F ..."
char serialDevice[128] = "/dev/ttyS0";
 
#define SLEEPTIME 2000L   // in us

// -------------------------------------------------
void sigalarm_handler(int sig)
{
  printf("Alarm signal received\n");
  closeSerialLine();
}   // end sigalarm_handler()

// -------------------------------------------------
void sigpipe_handler(int sig)
{
}   // end sigpipe_handler()

// -------------------------------------------------
// setup event handlers for CTRL-C etc
// (depends on underlying operating system)
void setup_event_handlers()
{
int    ret;
struct sigaction act;
  //
  // setup event handlers for CTRL-C etc.
  // The alarm handler terminates the program
  // This is necessary for HP-UX only!!!
  sigemptyset(&(act.sa_mask));
  sigaddset(&(act.sa_mask), SIGALRM);
  act.sa_handler = sigalarm_handler;
  sigaction(SIGALRM, &act, NULL);
  //
  sigemptyset(&(act.sa_mask));
  sigaddset(&(act.sa_mask), SIGPIPE);
  act.sa_handler = sigpipe_handler;
  sigaction(SIGPIPE, &act, NULL);
}   // end setup_event_handlers();

// ---------------------------------------------------------------------
unsigned long getEpochTime_ms(void)
{
  time_t   ltime;
  struct timeb tb;
  ltime = time(0);
  ftime(&tb);
  return (unsigned long) (ltime * 1000L + (int)tb.millitm);
}   // end getEpochTime_ms()

// -------------------------------
int closeSerialLine()
{
  int ret = 0;
  return ret;
}   // end closeSerialLine()

// ----------------------------------
// Returns: > 0 if data are available
//          < 0 upon error
int testDataAvaliable(int fd)
{
  int ret = 0;
  fd_set read_set;
  struct timeval tv;
  tv.tv_sec   = 0L;
  tv.tv_usec  = SLEEPTIME;   
  FD_ZERO(&read_set);
  FD_SET(fd, &read_set);
  // Hinweis: nicht 0 oder 1 einsetzen! stdin,stdout,stderr zaehlen mit!!!!
  // Unter Linux: FD_SETSIZE verwenden!!!
  ret = select(FD_SETSIZE, &read_set, NULL, NULL, &tv);
  if (ret > 0)
  {
    if (FD_ISSET(fd, &read_set))
      ret = 1;
    else
      ret = 0;
  }
  if (logging > 1)
    printf("testDataAvailable() ret=%d\n", ret);
  return ret;
}   // end testDataAvaliable()

// --------------------------------
// Returns: the number of bytes received
// Returns 0 if timeout
// Returns < 0 upon error
int readCommand(int fd, 
                char *cmdBuf, 
                int lBuf,
                unsigned long timeout)   // in ms
{
  int ret = 0;
  int nb = 0;
  int n;
  int lcount = 0;
  char readBuf[64];
  char ch;
  unsigned long t1, t2;
  //
  memset(cmdBuf, 0, lBuf);
  t1 = getEpochTime_ms();
  //
  while (1)
  {
    nb = testDataAvaliable(fd);
    if (nb <= 0)
    {
      usleep(SLEEPTIME);   // wait 2 ms
      if (lcount == 0)
      {
        t2 = getEpochTime_ms();
        if ((t2 - t1) >= timeout)
        {
          // nothing received during the given timeout time
          ret = 0;
          goto zurueck;
        }
      }
      continue;
    }   // end if nb <= 0
    //
    nb = read(fd, readBuf, sizeof(readBuf));
    if (logging > 1)
      printf("readCommand() nb=%d\n", nb);
    if (nb < 0)
    {
      if (logging)
        printf("readCommand() read error 1!\n");
      ret = -1;
      goto zurueck;
    }
    if (nb == 0)
    {
      // should not occur
      // If this happens the serial line must have been closed
      ret = -2;
      goto zurueck;
    }
    //
    // here nb > 0, i.e. characters received
    for (n=0; n < nb; n++)
    {
      ch = readBuf[n];
      if (logging)
      {
        if (ch <= ' ')
          printf("readCommand() char=0x%02u\n", readBuf[n] & 0xff);
        else
          printf("readCommand() char=%c\n", readBuf[n]);
      }
      if (lcount >= (lBuf - 1))
      {
        // line buffer overflow 
        lcount = 0;
        memset(cmdBuf, 0, lBuf);
        if (logging)
          printf("readCommand() Overflow!\n");
      }
      if (ch == '\r')
        // ignore carriage return
        continue;
      if (ch == '\n')
      {
         // a command is ready
         cmdBuf[lcount] = 0;
         ret = lcount;
         goto zurueck;
      }
      if (ch == '^')
      {
        // typical garbage after open()
        lcount = 0;
        memset(cmdBuf, 0, lBuf);
        continue;
      }
      // not (yet) LF received - put   
      // the received char into the command buffer
      cmdBuf[lcount++] = ch;
    }   // end for ...
  }   // end while (1)
  //
  zurueck:
  return ret;
}   // end readCommand()

// ---------------------------------
int openSerialLine(const char *pDevice)
{
  int fd;
  //
  fd = open(pDevice, O_RDONLY | O_SYNC | O_TRUNC);
  if (logging)
    printf("openSerialLine() for device=%s fd=%d\n", pDevice, fd);
  return fd;
}   // end openSerialLine()

// --------------------------------------------------
// this procedure will kill the process with name in var. play_cmd 
// It executes two commands via system() call 
int kill_process() 
{
  int logging = 0;
  FILE *ifd = NULL;
  unsigned count = 0;
  int globRet = 0;
  int ret = 0;
  static char cmd[256] = 
   "ps ax | grep testProc | cut -c-5 | head -n 1 >/tmp/pid.txt 2>/tmp/1.txt";
  int pid = 0;
  //
  strcpy(cmd, "ps ax | grep ");
  strcat(cmd, play_cmd);
  strcat(cmd, " | cut -c-5 | head -n 1 >/tmp/pid.txt 2>/tmp/1.txt");
  ret = system(cmd);
  if (logging)  printf("system() call 1 ret=%d\n", ret);
  if (ret != 0)
  {
    globRet = -1;
    goto zurueck;
  }
  ifd = fopen("/tmp/pid.txt", "r");
  if (ifd == NULL)
  {
    globRet = -2;
    goto zurueck;
  }
  if (logging) 
     printf("file /tmp/pid.txt OK\n");
  char buf[32];
  if (fgets(buf, sizeof(buf) - 1, ifd) == NULL)
  {
    globRet = -3;
    goto zurueck;
  }
  pid = atoi(buf);
  if (logging) 
    printf("pid=%d\n", pid);
  if (pid < 0)
  {
    globRet = -4;
    goto zurueck;
  }
  sprintf(cmd, "kill -9 %d 2>/tmp/2.txt", pid);
  ret = system(cmd); 
  if (logging) 
    printf("system() call 2 ret=%d\n", ret);
  if (ret != 0)
  {
    globRet = -5;
    goto zurueck;
  }
  //
  globRet = 0;
  //
  zurueck:
  if (ifd != NULL)
    fclose(ifd);
  return globRet;
}   // end kill_process()

// ------------------------------------------------------------------------
int play_wave_file(const char *parbuf)
{
  unsigned n;
  int ret = 0;
  char cmd[512];
  //
  sprintf(cmd, "%s", exe_dir);
  strcat(cmd, "/");
  strcat(cmd, play_cmd);
  strcat(cmd, " ");
  strcat(cmd, wave_dir);
  strcat(cmd, "/");
  strcat(cmd,  parbuf); 
  strcat(cmd , ".wav");
  strcat(cmd, " &");  // play in a background process
  //
  ret = system(cmd);
  //
  zurueck:
  return ret;
}   // end play_wave_file()

// ------------------------------------------------------------------------
int change_volume(const char *kind)
{
  int ret = 0;
  char cmd[512];
  //
  char cmdLow1[] = "amixer -c 0 sset  PCM 85% >dummy.txt";
  char cmdLow2[] = "amixer -c 0 sset  PCM 80% >dummy.txt";
  char cmdHigh1[] = "amixer -c 0 sset PCM 95% >dummy.txt";
  char cmdHigh2[] = "amixer -c 0 sset PCM 99% >dummy.txt";
  if (strstr(kind, "--") != NULL)
  {
    sprintf(cmd, "%s", cmdLow2);
    strcat(cmd, " &");  // submit as a background process
    ret = system(cmd);
    goto zurueck;
  }
  if (strstr(kind, "-") != NULL)
  {
    sprintf(cmd, "%s", cmdLow1);
    strcat(cmd, " &");  // submit as a background process
    ret = system(cmd);
    goto zurueck;
  }
  if (strstr(kind, "++") != NULL)
  {
    sprintf(cmd, "%s", cmdHigh2);
    strcat(cmd, " &");  // submit as a background process
    ret = system(cmd);
    goto zurueck;
  }
  if (strstr(kind, "+") != NULL)
  {
    sprintf(cmd, "%s", cmdHigh1);
    strcat(cmd, " &");  // submit as a background process
    ret = system(cmd);
    goto zurueck;
  }
  //
  zurueck:
  return ret;
}   // end change_volume()

// --------------------------------
int stop_player()
{
  int ret = 0;
  //
  printf("stopping the actual player...\n");
  ret = kill_process();
  //
  zurueck:
  return ret;
}   // end stop_player()

// -------------------------------
void usage()
{
  printf("usage: 433Mhz_server [-F device] [-v]\n");
  printf("Copyright Dr.E.Huckert V1.4 10-2017\n");
}   // end usage()

// -------------------------------
int main(int argc, char **argv)
{
    char buf[512];
    int ret = 0;
    int n, bytes_read;
    int serialFd = -1;
    //
    usage();
    // evaluate the command line
    for (n=0; n < argc; n++)
    {
       if (strcmp(argv[n], "-v") == 0)
         logging = 1;
       if (strcmp(argv[n], "-v2") == 0)
         logging = 2;
       if (strcmp(argv[n], "-F") == 0)
         strcpy(serialDevice, argv[n+1]);
    }
    //
    memset(buf, 0, sizeof(buf));
    serialFd = openSerialLine(serialDevice);
    if (serialFd < 0)
    {
       printf("ERROR: cannot open serial device=%s\n",
              serialDevice);
       ret = -1;
       goto zurueck;
    }
    //
    // read data from the serial device 
    // A command is an ASCII based string terminated by LF ('\n')
    unsigned count = 0;
    int handled = 0;
    //
    // endless loop here!
    while (1)
    {
      handled = 0;
      bytes_read = readCommand(serialFd, 
                               buf, 
                               sizeof(buf),
                               2000L);     // 2 secs timeout
      if (bytes_read == 0)
      {
        // timeout - nothing received
        continue;
      }
      if (logging)
        printf("main() bytes_read=%d\n", bytes_read);
      if (bytes_read < 0)
      {
        // the client has terminated the connection
        ret = -2;
        goto zurueck;
      }
      if (bytes_read == 0)
      {
         // must be a timeout - no bytes received
         usleep(2000L);  // sleep 2 ms
         continue;
      }
      if( bytes_read > 0 ) 
      {
        // a command is ready 
        // Note: characters like CR or LF are already eliminated
        buf[bytes_read] = 0;
        if (logging)
        {
          printf("RX #%04u: %s\n", count, buf);
          fflush(stdout);
        }
        count++;
        if (strncmp(buf, "STOP", 4) == 0)
        {
          handled = 1;
          stop_player();
          continue;
        }
        if (strstr(buf, "++") != NULL)
        {
          //  command for increasing the volume
          change_volume("++");
          handled = 1;
          continue;
        }
        if (strstr(buf, "--") != NULL)
        {
          //  command for lowering the volume
          change_volume("--");
          handled = 1;
          continue;
        }
        if (strstr(buf, "+") != NULL)
        {
          //  command for increasing the volume
          change_volume("+");
          handled = 1;
          continue;
        }
        if (strstr(buf, "-") != NULL)
        {
          //  command for lowering the volume
          change_volume("-");
          handled = 1;
          continue;
        }
        if (handled == 0) 
        {
          if (strncmp(buf, "HELLO", 4) == 0)
          {
            // ignore this message
            memset(buf, 0, sizeof(buf));
            continue;
          }
          if (strncmp(buf, "EXIT", 4) == 0)
          {
            // ignore this message
            memset(buf, 0, sizeof(buf));
            continue;
          }
          // this must be the name of a WAVE file to be played
          // simple plausibility check
          if ((buf[0] >= 'A') && (buf[0] <= 'z'))
          {
            play_wave_file(buf);
          }
          memset(buf, 0, sizeof(buf));
        }
      }  // end  if( bytes_read > 0 ) 
    }   // end while (1)
    //
    ret = 0;
    //
  zurueck:
    return ret;
}   // end main()
